use crate::config::Config;
use crate::patterns::Patterns;
use crate::Strip;
use bluer::Address;
use serde_derive::{Deserialize, Serialize};
use std::collections::HashMap;
use std::future::Future;
use std::str::FromStr;
use std::time::Duration;
use tokio::sync::watch::{Receiver, Sender};

#[derive(Serialize, Deserialize, Debug, Eq, Hash, PartialEq, Copy, Clone)]
pub struct Context<const N: usize> {
    pub(crate) pattern: Patterns<N>,
    pub period: Duration,
}

pub(crate) type DeviceSequence<const N: usize> = Vec<Context<N>>;

pub async fn runner_loop<const N: usize>(runner: DeviceSequence<N>, tx: Sender<Strip<N>>) {
    for mut context in runner {
        let mut strip = context.pattern.next().unwrap();
        let delay = context.period;
        println!("{:?}", delay);
        while tx.send(strip).is_ok() {
            if let Some(value) = context.pattern.next() {
                strip = value;
            } else {
                break;
            }
            tokio::time::sleep(delay).await;
        }
    }
}

#[derive(Default)]
pub struct Spectacle<const N: usize> {
    runners: HashMap<DeviceSequence<N>, (Sender<Strip<N>>, Receiver<Strip<N>>)>,
    channels: HashMap<Address, Receiver<Strip<N>>>,
    default_runner: Option<(DeviceSequence<N>, Sender<Strip<N>>)>,
    default_receiver: Option<Receiver<Strip<N>>>,
}

impl<const N: usize> Spectacle<N> {
    pub fn add(&mut self, mac_address: Address, runner: &DeviceSequence<N>) {
        if let Some((_, rx)) = self.runners.get(&runner as &DeviceSequence<N>) {
            self.channels.insert(mac_address, rx.clone());
        } else {
            let (tx, rx) = tokio::sync::watch::channel(Strip::<N>::default());
            self.runners.insert(runner.clone(), (tx, rx.clone()));
            self.channels.insert(mac_address, rx);
        }
    }

    pub fn add_default(&mut self, runner: DeviceSequence<N>) {
        let (tx, rx) = tokio::sync::watch::channel(Strip::<N>::default());
        // self.tasks.push(runner.runner_loop(tx));
        self.default_runner = Some((runner, tx));
        self.default_receiver = Some(rx);
    }

    pub fn get_channel(&self, mac_address: &Address) -> Option<Receiver<Strip<N>>> {
        let channel = self.channels.get(mac_address);
        match channel {
            None => {
                let d = &self.default_receiver;
                match d {
                    None => None,
                    Some(runner) => {
                        println!("default rx");
                        Some(runner.clone())
                    }
                }
            }
            Some(channel) => {
                println!("rx found ");
                Some(channel.clone())
            }
        }
    }

    pub fn run(&mut self) -> Vec<impl Future<Output = ()>> {
        let mut joinhandles = vec![];

        if let Some(_) = &self.default_runner {
            let d = std::mem::take(&mut self.default_runner).unwrap();
            joinhandles.push(runner_loop(d.0.clone(), d.1));
            println!("add default runner loop");
        }

        for (runner, (tx, _)) in self.runners.drain().take(1) {
            joinhandles.push(runner_loop(runner.clone(), tx));
            println!("add custom runner loop");
        }
        joinhandles
    }
}

impl<const N: usize> From<Config<N>> for Spectacle<N> {
    fn from(config: Config<N>) -> Self {
        let mut spectacle: Spectacle<N> = Spectacle::default();
        for device in config {
            if let Some(mac_addresses) = device.mac_addresses {
                for mac_address in mac_addresses {
                    let mac_address = Address::from_str(&*mac_address);
                    match mac_address {
                        Ok(address) => {
                            spectacle.add(address, &device.sequence);
                        }
                        Err(_) => {
                            println!("error reading mac address");
                        }
                    }
                }
            } else {
                spectacle.add_default(device.sequence);
            }
        }
        spectacle
    }
}
